@@ -38,6 +38,31 @@
 
 extern int yyparse(yyscan_t scanner, SearchCriteriaParseData *data);
 
+static uint32_t update_id = 0;
+
+static ServiceVariable *
+find_variable(const char *name)
+{
+    int c;
+    for (c = 0; contentdir_service_variables[c].name != NULL; c++) {
+        if (strcmp(contentdir_service_variables[c].name, name) == 0)
+            return &contentdir_service_variables[c];
+    }
+    assert(0); /* Shouldn't get here */
+}
+
+void
+bump_update_id(void)
+{
+    ServiceVariable *var;
+
+    var = find_variable("SystemUpdateID");
+    update_id++;
+    free(var->value);
+    var->value = xstrdup(uint32_str(update_id));
+    notify_change(CONTENTDIR_SERVICE_ID, var);
+}
+
 void
 free_search_criteria(SearchCriteria *criteria)
 {
@@ -85,7 +110,7 @@ contentdir_get_sort_capabilities(ActionEvent *event)
 static bool
 contentdir_get_system_update_id(ActionEvent *event)
 {
-    upnp_add_response(event, "Id", "0");
+    upnp_add_response(event, "Id", uint32_str(update_id));
     return event->status;
 }
 
@@ -95,7 +120,7 @@ append_escaped_xml(StrBuf *out, const char *entity, const char *value)
     char *str;
 
     str = xsgmlescape(value);
-    strbuf_appendf(out, "<%s>%s</%s>", entity, value, entity);
+    strbuf_appendf(out, "<%s>%s</%s>", entity, str, entity);
     free(str);
 }
 
@@ -127,6 +152,15 @@ get_entry_property(Entry *entry, char *property)
      * such as @refID, @id, @parentID or @protocolInfo?
      */
 
+    /* See the ContentDirectory spec for an example where @id is used. */
+    if (strcmp(property, "id") == 0 || strcmp(property, "@id") == 0) {
+        char *t1, *t2;
+        t1 = xasprintf("%" PRId32, entry->id); /* " */
+        t2 = convert_string(t1);
+        free(t1);
+        return t2;
+    }
+
     if (strcmp(property, "dc:title") == 0) {
 	if (detail != NULL && detail->data.tag.title != NULL && strcmp(detail->data.tag.title, "") != 0)
 	    return convert_string(detail->data.tag.title);
@@ -142,17 +176,43 @@ get_entry_property(Entry *entry, char *property)
     }
 
     if (detail != NULL) {
-	if (strcmp(property, "upnp:artist") == 0)
+        /* XXX: Accept dc:creator as a synonym to upnp:artist? */
+	if (strcmp(property, "upnp:artist") == 0 && detail->data.tag.artist != NULL)
 	    return convert_string(detail->data.tag.artist);
-	if (strcmp(property, "upnp:album") == 0)
+	if (strcmp(property, "upnp:album") == 0 && detail->data.tag.album != NULL)
 	    return convert_string(detail->data.tag.album);
-	if (strcmp(property, "upnp:genre") == 0)
+	if (strcmp(property, "upnp:genre") == 0 && detail->data.tag.genre != NULL)
 	    return convert_string(detail->data.tag.genre);
     }
 
     return NULL;
 }
 
+static bool
+filter_includes_any(const char *filter, ...)
+{
+    va_list args;
+    const char *field;
+    bool result = false;
+
+    if (strcmp(filter, "*") == 0)
+        return true;
+
+    va_start(args, filter);
+    for (;;) {
+        field = va_arg(args, const char *);
+        if (field == NULL)
+            break;
+        if (string_in_csv(filter, ',', field)) {
+            result = true;
+            break;
+        }
+    }
+    va_end(args);
+
+    return result;
+}
+
 static bool
 filter_includes(const char *filter, const char *field)
 {
@@ -183,14 +243,14 @@ append_entry(StrBuf *result, Entry *entry, const char *filter)
 		"<item id=\"%" PRIi32 "\" parentID=\"%" PRIi32 "\" restricted=\"true\">",
 		entry->id,
 		entry->parent);
-        if (filter_includes(filter, "res") || filter_includes(filter, "@protocolInfo") || filter_includes(filter, "@size")) {
+        if (filter_includes_any(filter, "res", "@protocolInfo", "@size", "res@protocolInfo", "res@size", NULL)) {
             detail = get_entry_detail(entry, DETAIL_FILE);
             if (detail != NULL) {
                 strbuf_append(result, "<res");
-                if (filter_includes(filter, "res") || filter_includes(filter, "@protocolInfo"))
+                if (filter_includes_any(filter, "res", "@protocolInfo", "res@protocolInfo", NULL))
                     strbuf_appendf(result, " protocolInfo=\"http-get:*:%s:*\"", detail->data.file.mime_type);
                 /*if (detail->data.file.size != 0)*/
-                if (filter_includes(filter, "res") || filter_includes(filter, "@size"))
+                if (filter_includes_any(filter, "res", "@size", "res@size", NULL))
                     strbuf_appendf(result, " size=\"%" PRIu32 "\"", detail->data.file.size);
                 strbuf_appendf(result, ">http://%s:%d/files/%d</res>",
                         UpnpGetServerIpAddress(),
@@ -487,20 +547,6 @@ contentdir_search(ActionEvent *event)
     if (count == 0)
         count = UINT32_MAX;
 
-    /* Check ContainerID argument */
-    entry = get_entry_by_id(id);
-    if (entry == NULL) {
-        upnp_set_error(event, UPNP_CONTENTDIR_E_NO_CONTAINER,
-            _("No such container"));
-	return false;
-    }
-    detail = get_entry_detail(entry, DETAIL_CHILDREN);
-    if (detail == NULL) {
-        upnp_set_error(event, UPNP_CONTENTDIR_E_NO_CONTAINER,
-            _("Not a container"));
-	return false;
-    }
-
     /* SortCriteria not supported at the moment */
     if (strcmp(sort_criteria, "") != 0) {
         upnp_set_error(event, UPNP_CONTENTDIR_E_BAD_SORT_CRITERIA,
@@ -520,6 +566,23 @@ contentdir_search(ActionEvent *event)
 	dump_search_expr(criteria->expr, 2);
     }
 
+    /* Check ContainerID argument */
+    lock_metadata();
+    entry = get_entry_by_id(id);
+    if (entry == NULL) {
+        upnp_set_error(event, UPNP_CONTENTDIR_E_NO_CONTAINER,
+            _("No such container"));
+        unlock_metadata();
+	return false;
+    }
+    detail = get_entry_detail(entry, DETAIL_CHILDREN);
+    if (detail == NULL) {
+        upnp_set_error(event, UPNP_CONTENTDIR_E_NO_CONTAINER,
+            _("Not a container"));
+        unlock_metadata();
+	return false;
+    }
+
     /* Do the actual searching */
     result = strbuf_new();
     begin_result(result);
@@ -530,8 +593,9 @@ contentdir_search(ActionEvent *event)
     upnp_add_response(event, "Result", strbuf_buffer(result));
     upnp_add_response(event, "NumberReturned", int32_str(match_count));
     upnp_add_response(event, "TotalMatches", "0");
-    upnp_add_response(event, "UpdateID", "0");
+    upnp_add_response(event, "UpdateID", uint32_str(id == 0 ? update_id : 0));
     strbuf_free(result);
+    unlock_metadata();
 
     return true;
 }
@@ -587,11 +651,13 @@ contentdir_browse(ActionEvent *event)
 	return false;
     }
 
+    lock_metadata();
     if (metadata) {
         entry = get_entry_by_id(id);
 	if (entry == NULL) {
             upnp_set_error(event, UPNP_CONTENTDIR_E_NO_OBJECT,
                 _("No such object"));
+            unlock_metadata();
 	    return false;
 	}
 
@@ -615,6 +681,7 @@ contentdir_browse(ActionEvent *event)
 	if (entry == NULL) {
             upnp_set_error(event, UPNP_CONTENTDIR_E_NO_OBJECT,
                 _("No such object"));
+            unlock_metadata();
 	    return false;
 	}
 
@@ -622,6 +689,7 @@ contentdir_browse(ActionEvent *event)
         if (detail == NULL) {
             upnp_set_error(event, UPNP_SOAP_E_INVALID_ARGS,
                 _("BrowseDirectChildren only possible on containers"));
+            unlock_metadata();
             return false;
         }
 
@@ -643,33 +711,33 @@ contentdir_browse(ActionEvent *event)
         strbuf_free(result);
     }
 
-    upnp_add_response(event, "UpdateID", "0");
+    upnp_add_response(event, "UpdateID", uint32_str(id == 0 ? update_id : 0));
+    unlock_metadata();
 
     return event->status;
 }
 
-/*
+void
+init_contentdir(void)
+{
+    find_variable("SystemUpdateID")->value = xstrdup(uint32_str(update_id));
+}
+
+void
+finish_contentdir(void)
+{
+    free(find_variable("SystemUpdateID"));
+}
+
+/* XXX: In the future, ServiceVariable should allow different
+ * data types (and point to real variables, e.g. {"SystemUpdateID", UINT32_TYPE, &update_id}).
+ */
 ServiceVariable contentdir_service_variables[] = {
-  { "A_ARG_TYPE_ObjectID"
-  { "A_ARG_TYPE_Result"
-//{ "A_ARG_TYPE_SearchCriteria"
-  { "A_ARG_TYPE_BrowseFlag"
-  { "A_ARG_TYPE_Filter"
-  { "A_ARG_TYPE_SortCriteria"
-  { "A_ARG_TYPE_Index"
-  { "A_ARG_TYPE_Count"
-  { "A_ARG_TYPE_UpdateID"
-//{ "A_ARG_Type_TransferID"
-//{ "A_ARG_Type_TransferLength"
-//{ "A_ARG_Type_TransferTotal"
-//{ "A_ARG_TYPE_TagValueList"
-//{ "A_ARG_TYPE_URI"
-  { "SearchCapabilities"
-  { "SortCapabilities"
-  { "SystemUpdateID"
-//{ "ContainerUpdateIDs"
+  /*{ "TransferIDs", NULL },*/
+  { "SystemUpdateID", NULL },
+  /*{ "ContainerUpdateIDs", NULL },*/
+  { 0, }
 };
-*/
 
 ServiceAction contentdir_service_actions[] = {
   { "GetSearchCapabilities", contentdir_get_search_capabilities },
@@ -684,5 +752,5 @@ ServiceAction contentdir_service_actions[] = {
 /*{ "GetTransferProgress", contentdir_get_transfer_progress },*/
 /*{ "DeleteResource", contentdir_delete_resource },*/
 /*{ "CreateReference", contentdir_create_reference },*/
-  { NULL, NULL }
+  { 0, }
 };
