@@ -32,6 +32,7 @@
 #include "strbuf.h"
 
 typedef struct _ServiceAction ServiceAction;
+typedef struct _ServiceVariable ServiceVariable;
 typedef struct _Service Service;
 typedef struct _ActionEvent ActionEvent;
 typedef struct _EntryChildren EntryChildren;
@@ -58,10 +59,16 @@ struct _ServiceAction {
     bool (*function)(ActionEvent *);
 };
 
+struct _ServiceVariable {
+    char *name;
+    char *value;
+};
+
 struct _Service {
     char *id;
     char *type;
     ServiceAction *actions;
+    ServiceVariable *variables;
 };
 
 struct _ActionEvent {
@@ -184,8 +191,12 @@ extern char *file_types;
 #ifdef HAVE_ID3LIB
 extern bool id3_enabled;
 #endif
-bool init_metadata(char **pathv, int pathc);
+void init_metadata(void);
+bool scan_entries(char **pathv, int pathc);
+void clear_entries(void);
 void finish_metadata(void);
+void lock_metadata(void);
+void unlock_metadata(void);
 Entry *get_entry_by_id(uint32_t id);
 EntryDetail *get_entry_detail(Entry *entry, EntryDetailType type);
 bool has_entry_detail(Entry *entry, EntryDetailType type);
@@ -220,6 +231,7 @@ char *upnp_get_string(ActionEvent *event, const char *item);
 uint32_t upnp_get_ui4(ActionEvent *event, const char *key);
 bool upnp_add_response(ActionEvent *event, char *key, const char *value);
 int32_t upnp_get_i4(ActionEvent *event, const char *key);
+void notify_change(const char *service_id, ServiceVariable *var);
 
 /* webserver.c */
 extern struct UpnpVirtualDirCallbacks virtual_dir_callbacks;
@@ -232,12 +244,17 @@ ssize_t http_read(HTTPResult *result, void *buf, size_t count);
 bool parse_http_date(const char *s, time_t *timeptr);
 
 /* contentdir.c */
+#define CONTENTDIR_SERVICE_ID "urn:upnp-org:serviceId:ContentDirectory"
 extern ServiceAction contentdir_service_actions[];
+extern ServiceVariable contentdir_service_variables[];
+void bump_update_id(void);
 void free_search_criteria(SearchCriteria *criteria);
 void free_search_expr(SearchExpr *expr);
 
 /* connectmgr.c */
+#define CONNECTMGR_SERVICE_ID "urn:upnp-org:serviceId:ConnectionManager"
 extern ServiceAction connectmgr_service_actions[];
+extern ServiceVariable connectmgr_service_variables[];
 
 /* url.c */
 void sgmlescape(const char *str, char *target, uint32_t *length);
